
/* Chrysalide - Outil d'analyse de fichiers binaires
 * undefined.c - instructions au comportement non défini
 *
 * Copyright (C) 2016 Cyrille Bagard
 *
 *  This file is part of Chrysalide.
 *
 *  Chrysalide is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  Chrysalide is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
 */


#include "undefined.h"


#include <assert.h>


#include <i18n.h>


#include "instruction-int.h"



/* Définition générique d'une instruction au comportement non défini (instance) */
struct _GUndefInstruction
{
    GArchInstruction parent;                /* A laisser en premier        */

    InstrBehaviorStatus status;             /* Conséquences réelles        */

};

/* Définition générique d'une instruction au comportement non défini (classe) */
struct _GUndefInstructionClass
{
    GArchInstructionClass parent;           /* A laisser en premier        */

};


/* Initialise la classe des instructions non définies. */
static void g_undef_instruction_class_init(GUndefInstructionClass *);

/* Initialise une instance d'instruction non définie. */
static void g_undef_instruction_init(GUndefInstruction *);

/* Supprime toutes les références externes. */
static void g_undef_instruction_dispose(GUndefInstruction *);

/* Procède à la libération totale de la mémoire. */
static void g_undef_instruction_finalize(GUndefInstruction *);

/* Indique l'encodage d'une instruction de façon détaillée. */
static const char *g_undef_instruction_get_encoding(const GUndefInstruction *);

/* Fournit le nom humain de l'instruction manipulée. */
static const char *g_undef_instruction_get_keyword(const GUndefInstruction *, AsmSyntax);

/* Ajoute à un tampon GLib le contenu de l'instance spécifiée. */
static GBufferLine *g_undef_instruction_print(GUndefInstruction *, GCodeBuffer *, MemoryDataSize, const GBinContent *, AsmSyntax);



/* ---------------------------------------------------------------------------------- */
/*                           INSTRUCTION INCONNUE / DONNEES                           */
/* ---------------------------------------------------------------------------------- */


/* Indique le type défini pour une instruction au comportement non défini. */
G_DEFINE_TYPE(GUndefInstruction, g_undef_instruction, G_TYPE_ARCH_INSTRUCTION);


/******************************************************************************
*                                                                             *
*  Paramètres  : klass = classe à initialiser.                                *
*                                                                             *
*  Description : Initialise la classe des instructions non définies.          *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_undef_instruction_class_init(GUndefInstructionClass *klass)
{
    GObjectClass *object;                   /* Autre version de la classe  */
    GArchInstructionClass *instr;           /* Encore une autre vision...  */

    object = G_OBJECT_CLASS(klass);

    object->dispose = (GObjectFinalizeFunc/* ! */)g_undef_instruction_dispose;
    object->finalize = (GObjectFinalizeFunc)g_undef_instruction_finalize;

    instr = G_ARCH_INSTRUCTION_CLASS(klass);

    instr->get_encoding = (get_instruction_encoding_fc)g_undef_instruction_get_encoding;
    instr->get_keyword = (get_instruction_keyword_fc)g_undef_instruction_get_keyword;
    instr->print = (print_instruction_fc)g_undef_instruction_print;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : instr = instance à initialiser.                              *
*                                                                             *
*  Description : Initialise une instance d'instruction non définie.           *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_undef_instruction_init(GUndefInstruction *instr)
{

}


/******************************************************************************
*                                                                             *
*  Paramètres  : instr = instance d'objet GLib à traiter.                     *
*                                                                             *
*  Description : Supprime toutes les références externes.                     *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_undef_instruction_dispose(GUndefInstruction *instr)
{
    G_OBJECT_CLASS(g_undef_instruction_parent_class)->dispose(G_OBJECT(instr));

}


/******************************************************************************
*                                                                             *
*  Paramètres  : instr = instance d'objet GLib à traiter.                     *
*                                                                             *
*  Description : Procède à la libération totale de la mémoire.                *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_undef_instruction_finalize(GUndefInstruction *instr)
{
    G_OBJECT_CLASS(g_undef_instruction_parent_class)->finalize(G_OBJECT(instr));

}


/******************************************************************************
*                                                                             *
*  Paramètres  : status = état réel du CPU après le passage de l'instruction. *
*                                                                             *
*  Description : Crée une instruction au comportement nominalement indéfini.  *
*                                                                             *
*  Retour      : Instruction mise en place.                                   *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

GArchInstruction *g_undef_instruction_new(InstrBehaviorStatus status)
{
    GArchInstruction *result;               /* Instruction à retourner     */

    result = g_object_new(G_TYPE_UNDEF_INSTRUCTION, NULL);

    G_UNDEF_INSTRUCTION(result)->status = status;

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : instr = instruction quelconque à consulter.                  *
*                                                                             *
*  Description : Indique l'encodage d'une instruction de façon détaillée.     *
*                                                                             *
*  Retour      : Description humaine de l'encodage utilisé.                   *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static const char *g_undef_instruction_get_encoding(const GUndefInstruction *instr)
{
    const char *result;                     /* Description à retourner     */

    result = _("Undefined");

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : instr  = instruction d'assemblage à consulter.               *
*                syntax = type de représentation demandée.                    *
*                                                                             *
*  Description : Fournit le nom humain de l'instruction manipulée.            *
*                                                                             *
*  Retour      : Mot clef de bas niveau.                                      *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

const char *g_undef_instruction_get_keyword(const GUndefInstruction *instr, AsmSyntax syntax)
{
    const char *result;                     /* Désignation à retourner     */

    switch (instr->status)
    {
        case IBS_NOP:
            result = "nop";
            break;

        case IBS_UNDEFINED:
            result = "undefined";
            break;

        case IBS_UNPREDICTABLE:
            result = "unpredictable";
            break;

        default:
            assert(false);
            break;

    }

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : instr  = instruction d'assemblage à représenter.             *
*                buffer = espace où placer ledit contenu.                     *
*                msize   = taille idéale des positions et adresses;           *
*                content = contenu binaire global à venir lire.               *
*                syntax = type de représentation demandée.                    *
*                                                                             *
*  Description : Ajoute à un tampon GLib le contenu de l'instance spécifiée.  *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static GBufferLine *g_undef_instruction_print(GUndefInstruction *instr, GCodeBuffer *buffer, MemoryDataSize msize, const GBinContent *content, AsmSyntax syntax)
{
    GBufferLine *result;                    /* Ligne de destination        */
    GArchInstruction *base;                 /* Version de base             */
    const char *key;                        /* Mot clef principal          */
    size_t klen;                            /* Taille de ce mot clef       */

    base = G_ARCH_INSTRUCTION(instr);

    result = g_code_buffer_prepare_new_line(buffer, &base->range);

    g_buffer_line_add_flag(result, BLF_HAS_CODE);

    g_buffer_line_fill_for_instr(result, msize/* TODO ! */, msize, content, base->max_displayed_len);

    /* Instruction proprement dite */

    key = g_arch_instruction_get_keyword(base, syntax);
    klen = strlen(key);

    g_buffer_line_append_text(result, BLC_ASSEMBLY_HEAD, key, klen, RTT_ERROR, NULL);

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : instr  = instruction à consulter.                            *
*                                                                             *
*  Description : Indique le type de conséquences réél de l'instruction.       *
*                                                                             *
*  Retour      : Etat réel du CPU après l'exécution de l'instruction.         *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

InstrBehaviorStatus g_undef_instruction_get_status(const GUndefInstruction *instr)
{
    return instr->status;

}
